@@ -1,7 +1,7 @@
 /*
  * ipc.h - Interprocess communication routines. Handlers read and write.
  *
- * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -10,12 +10,14 @@
 #ifndef NSD_IPC_H
 #define NSD_IPC_H
 
-#include "config.h"
 #include "netio.h"
 struct buffer;
 struct nsd;
 struct nsd_child;
 struct xfrd_tcp;
+struct xfrd_state;
+struct nsdst;
+struct event;
 
 /*
  * Data for the server_main IPC handler 
@@ -35,10 +37,6 @@ struct main_ipc_handler_data
 	uint16_t	total_bytes;
 	uint32_t	acl_num;
 	int32_t		acl_xfr;
-	
-	/* writing data, connection and state */
-	uint8_t		busy_writing_zone_state;
-	struct xfrd_tcp	*write_conn;
 };
 
 /*
@@ -78,17 +76,21 @@ void parent_handle_child_command(netio_type *netio,
  * Routine used by server_child.
  * Handle a command received from the parent process.
  */
-void child_handle_parent_command(netio_type *netio,
-	netio_handler_type *handler, netio_event_types_type event_types);
+void child_handle_parent_command(int fd, short event, void* arg);
 
 /*
  * Routine used by xfrd
  * Handle interprocess communication with parent process, read and write.
  */
-void xfrd_handle_ipc(netio_type *netio,
-	netio_handler_type *handler, netio_event_types_type event_types);
+void xfrd_handle_ipc(int fd, short event, void* arg);
 
 /* check if all children have exited in an orderly fashion and set mode */
 void parent_check_all_children_exited(struct nsd* nsd);
 
+/** add stats to total */
+void stats_add(struct nsdst* total, struct nsdst* s);
+
+/** set event to listen to given mode, no timeout, must be added already */
+void ipc_xfrd_set_listening(struct xfrd_state* xfrd, short mode);
+
 #endif /* NSD_IPC_H */
