@@ -1,7 +1,7 @@
 /*
  * zonec.h -- zone compiler.
  *
- * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -71,6 +71,7 @@ extern domain_type *error_domain;
 
 int yyparse(void);
 int yylex(void);
+int yylex_destroy(void);
 /*int yyerror(const char *s);*/
 void yyrestart(FILE *);
 
@@ -79,6 +80,9 @@ void zc_warning_prev_line(const char *fmt, ...) ATTR_FORMAT(printf, 1, 2);
 void zc_error(const char *fmt, ...) ATTR_FORMAT(printf, 1, 2);
 void zc_error_prev_line(const char *fmt, ...) ATTR_FORMAT(printf, 1, 2);
 
+void parser_push_stringbuf(char* str);
+void parser_pop_stringbuf(void);
+
 int process_rr(void);
 uint16_t *zparser_conv_hex(region_type *region, const char *hex, size_t len);
 uint16_t *zparser_conv_hex_length(region_type *region, const char *hex, size_t len);
@@ -92,10 +96,10 @@ uint16_t *zparser_conv_byte(region_type *region, const char *text);
 uint16_t *zparser_conv_a(region_type *region, const char *text);
 uint16_t *zparser_conv_aaaa(region_type *region, const char *text);
 uint16_t *zparser_conv_ilnp64(region_type *region, const char *text);
-#ifdef DRAFT_RRTYPES
 uint16_t *zparser_conv_eui(region_type *region, const char *text, size_t len);
-#endif
 uint16_t *zparser_conv_text(region_type *region, const char *text, size_t len);
+uint16_t *zparser_conv_long_text(region_type *region, const char *text, size_t len);
+uint16_t *zparser_conv_tag(region_type *region, const char *text, size_t len);
 uint16_t *zparser_conv_dns_name(region_type *region, const uint8_t* name, size_t len);
 uint16_t *zparser_conv_b32(region_type *region, const char *b32);
 uint16_t *zparser_conv_b64(region_type *region, const char *b64);
@@ -126,4 +130,17 @@ zparser_type *zparser_create(region_type *region, region_type *rr_region,
 void zparser_init(const char *filename, uint32_t ttl, uint16_t klass,
 		  const dname_type *origin);
 
+/* parser start and stop to parse a zone */
+void zonec_setup_parser(namedb_type* db);
+void zonec_desetup_parser(void);
+/* parse a zone into memory. name is origin. zonefile is file to read.
+ * returns number of errors; failure may have read a partial zone */
+unsigned int zonec_read(const char *name, const char *zonefile, zone_type* zone);
+/* parse a string into the region. and with given domaintable. global parser
+ * is restored afterwards. zone needs apex set. returns last domain name
+ * parsed and the number rrs parse. return number of errors, 0 is success.
+ * The string must end with a newline after the RR. */
+int zonec_parse_string(region_type* region, domain_table_type* domains,
+	zone_type* zone, char* str, domain_type** parsed, int* num_rrs);
+
 #endif /* _ZONEC_H_ */
